@@ -1,4 +1,7 @@
 package CHI::Driver;
+{
+  $CHI::Driver::VERSION = '0.58';
+}
 use Carp;
 use CHI::CacheObject;
 use CHI::Constants qw(CHI_Max_Time);
@@ -9,15 +12,17 @@ use CHI::Driver::Role::IsSubcache;
 use CHI::Driver::Role::Universal;
 use CHI::Serializer::Storable;
 use CHI::Serializer::JSON;
-use CHI::Util qw(has_moose_class parse_duration);
-use CHI::Types;
+use CHI::Util qw(parse_duration);
+use CHI::Types qw(:all);
 use Digest::MD5;
 use Encode;
+use Hash::MoreUtils qw(slice_grep);
 use Log::Any qw($log);
-use Moose;
-use Moose::Util::TypeConstraints;
+use Moo;
+use MooX::Types::MooseLike::Base qw(:all);
 use Scalar::Util qw(blessed);
 use Time::Duration;
+use Time::HiRes qw(gettimeofday);
 use strict;
 use warnings;
 
@@ -25,58 +30,143 @@ my $default_serializer     = CHI::Serializer::Storable->new();
 my $default_key_serializer = CHI::Serializer::JSON->new();
 my $default_key_digester   = Digest::MD5->new();
 
-has 'chi_root_class'     => ( is => 'ro' );
-has 'constructor_params' => ( is => 'ro', init_arg => undef );
-has 'driver_class'       => ( is => 'ro' );
-has 'expires_at'         => ( is => 'rw', default => CHI_Max_Time );
-has 'expires_in' => ( is => 'rw', isa => 'CHI::Types::Duration', coerce => 1 );
-has 'expires_variance' => ( is => 'rw', default => 0.0 );
-has 'has_subcaches' =>
-  ( is => 'ro', isa => 'Bool', default => undef, init_arg => undef );
-has 'is_size_aware' => ( is => 'ro', isa => 'Bool', default => undef );
-has 'is_subcache'   => ( is => 'ro', isa => 'Bool', default => undef );
-has 'key_digester'  => (
-    is      => 'ro',
-    isa     => 'CHI::Types::Digester',
-    coerce  => 1,
-    default => sub { $default_key_digester }
-);
-has 'key_serializer' => (
-    is      => 'ro',
-    isa     => 'CHI::Types::Serializer',
-    coerce  => 1,
-    default => sub { $default_key_serializer }
-);
-has 'label'           => ( is => 'rw', lazy_build => 1 );
-has 'max_build_depth' => ( is => 'ro', default    => 8 );
-has 'max_key_length' => ( is => 'ro', isa        => 'Int', default => 1 << 31 );
-has 'metacache'      => ( is => 'ro', lazy_build => 1 );
-has 'namespace' => ( is => 'ro', isa => 'Str', default => 'Default' );
-has 'on_get_error' =>
-  ( is => 'rw', isa => 'CHI::Types::OnError', default => 'log' );
-has 'on_set_error' =>
-  ( is => 'rw', isa => 'CHI::Types::OnError', default => 'log' );
-has 'serializer' => (
-    is      => 'ro',
-    isa     => 'CHI::Types::Serializer',
-    coerce  => 1,
-    default => sub { $default_serializer }
-);
-has 'short_driver_name' => ( is => 'ro', lazy_build => 1 );
+my @common_params;
+{
+    my %attr = (
+        chi_root_class => {
+            is => 'ro',
+        },
+        compress_threshold => {
+            is => 'ro',
+            isa => Int,
+        },
+        constructor_params => {
+            is => 'ro',
+            init_arg => undef,
+        },
+        driver_class => {
+            is => 'ro',
+        },
+        expires_at => {
+            is => 'rw',
+            default => sub { CHI_Max_Time },
+        },
+        expires_in => { is => 'rw',
+            isa => Duration,
+            coerce => \&to_Duration,
+        },
+        expires_on_backend => {
+            is => 'ro',
+            isa => Num,
+            default => sub { 0 },
+        },
+        expires_variance => {
+            is => 'rw',
+            isa => Num,
+            default => sub { 0 },
+        },
+        has_subcaches => {
+            is => 'lazy',
+            isa => Bool,
+            init_arg => undef,
+        },
+        is_size_aware => {
+            is => 'ro',
+            isa => Bool,
+        },
+        is_subcache => {
+            is => 'ro',
+            isa => Bool,
+        },
+        key_digester => {
+            is => 'ro',
+            isa => Digester,
+            coerce => \&to_Digester,
+            default => sub { $default_key_digester },
+        },
+        key_serializer => {
+            is => 'ro',
+            isa => Serializer,
+            coerce => \&to_Serializer,
+            default => sub { $default_key_serializer },
+        },
+        label => {
+            is => 'rw',
+            lazy => 1,
+            builder => 1,
+            clearer => 1,
+            predicate => 1,
+        },
+        max_build_depth => {
+            is => 'ro',
+            default => sub { 8 },
+        },
+        max_key_length => {
+            is => 'ro',
+            isa => Int,
+            default => sub { 1 << 31 },
+        },
+        metacache => {
+            is => 'lazy',
+            clearer => 1,
+            predicate => 1,
+        },
+        namespace => {
+            is => 'ro',
+            isa => Str,
+            default => sub { 'Default' },
+        },
+        on_get_error => {
+            is => 'rw',
+            isa => OnError,
+            default => sub { 'log' },
+        },
+        on_set_error => {
+            is => 'rw',
+            isa => OnError,
+            default => sub { 'log' },
+        },
+        serializer => {
+            is => 'ro',
+            isa => Serializer,
+            coerce => \&to_Serializer,
+            default => sub { $default_serializer },
+        },
+        short_driver_name => {
+            is => 'lazy',
+            clearer => 1,
+            predicate => 1,
+        },
+        storage => {
+            is => 'ro',
+        },
+    );
+    push @common_params, keys %attr;
+    for my $attr ( keys %attr ) {
+        has $attr => %{$attr{$attr}};
+    }
+}
+
+sub _build_has_subcaches { undef }
 
 # These methods must be implemented by subclass
 foreach my $method (qw(fetch store remove get_keys get_namespaces)) {
-    __PACKAGE__->meta->add_method( $method =>
-          sub { die "method '$method' must be implemented by subclass" } );
+    no strict 'refs';
+    *{$method} = sub { die "method '$method' must be implemented by subclass" };
 }
 
-__PACKAGE__->meta->make_immutable();
-
 # Given a hash of params, return the subset that are not in CHI's common parameters.
 #
-my @common_params = map { $_->meta->get_attribute_list() } (
-    'CHI::Driver',                    'CHI::Driver::Role::HasSubcaches',
-    'CHI::Driver::Role::IsSizeAware', 'CHI::Driver::Role::IsSubcache'
+push @common_params, qw(
+    discard_policy
+    discard_timeout
+    l1_cache
+    max_size
+    max_size_reduction_factor
+    mirror_cache
+    parent_cache
+    subcache_type
+    subcaches
 );
 my %common_params = map { ( $_, 1 ) } @common_params;
 
@@ -93,15 +183,20 @@ sub declare_unsupported_methods {
     my ( $class, @methods ) = @_;
 
     foreach my $method (@methods) {
-        $class->meta->add_method( $method =>
-              sub { croak "method '$method' not supported by '$class'" } );
+        no strict 'refs';
+        *{"$class\::$method"} = sub { croak "method '$method' not supported by '$class'" };
     }
 }
 
+sub cache_object_class { 'CHI::CacheObject' }
+
 # To override time() for testing - must be writable in a dynamically scoped way from tests
 our $Test_Time;    ## no critic (ProhibitPackageVars)
 our $Build_Depth = 0;    ## no critic (ProhibitPackageVars)
 
+sub valid_get_options { qw(expire_if busy_lock) }
+sub valid_set_options { qw(expires_at expires_in expires_variance) }
+
 sub BUILD {
     my ( $self, $params ) = @_;
 
@@ -160,28 +255,29 @@ sub get {
     my ( $self, $key, %params ) = @_;
 
     croak "must specify key" unless defined($key);
-    $key = $self->transform_key($key);
-    my $ns_stats = $self->{ns_stats};
+    my $ns_stats     = $self->{ns_stats};
+    my $log_is_debug = $log->is_debug;
+    my $measure_time = defined($ns_stats) || $log_is_debug;
+    my ( $start_time, $elapsed_time );
 
     # Fetch cache object
     #
-    my $data = $params{data};
-    if ( !defined $data ) {
-        $data = eval { $self->fetch($key) };
-        if ( my $error = $@ ) {
-            $ns_stats->{'get_errors'}++ if defined($ns_stats);
-            $self->_handle_get_error( $error, $key );
-            return undef;
-        }
+    $start_time = gettimeofday() if $measure_time;
+    my $obj = eval { $params{obj} || $self->get_object($key) };
+    $elapsed_time = ( gettimeofday() - $start_time ) * 1000 if $measure_time;
+    if ( my $error = $@ ) {
+        $ns_stats->{'get_errors'}++ if defined($ns_stats);
+        $self->_handle_get_error( $error, $key );
+        return undef;
     }
-
-    if ( !defined $data ) {
-        $ns_stats->{'absent_misses'}++ if defined($ns_stats);
-        $self->_log_get_result( $log, "MISS (not in cache)", $key )
-          if $log->is_debug;
+    if ( !defined $obj ) {
+        $self->_record_get_stats( 'absent_misses', $elapsed_time )
+          if defined($ns_stats);
+        $self->_log_get_result( $log, "MISS (not in cache)",
+            $key, $elapsed_time )
+          if $log_is_debug;
         return undef;
     }
-    my $obj = $self->unpack_from_data( $key, $data );
     if ( defined( my $obj_ref = $params{obj_ref} ) ) {
         $$obj_ref = $obj;
     }
@@ -191,9 +287,10 @@ sub get {
     my $is_expired = $obj->is_expired()
       || ( defined( $params{expire_if} ) && $params{expire_if}->($obj) );
     if ($is_expired) {
-        $ns_stats->{'expired_misses'}++ if defined($ns_stats);
-        $self->_log_get_result( $log, "MISS (expired)", $key )
-          if $log->is_debug;
+        $self->_record_get_stats( 'expired_misses', $elapsed_time )
+          if defined($ns_stats);
+        $self->_log_get_result( $log, "MISS (expired)", $key, $elapsed_time )
+          if $log_is_debug;
 
         # If busy_lock value provided, set a new "temporary" expiration time that many
         # seconds forward before returning undef
@@ -209,15 +306,22 @@ sub get {
         return undef;
     }
 
-    $ns_stats->{'hits'}++ if defined($ns_stats);
-    $self->_log_get_result( $log, "HIT", $key ) if $log->is_debug;
+    $self->_record_get_stats( 'hits', $elapsed_time ) if defined($ns_stats);
+    $self->_log_get_result( $log, "HIT", $key, $elapsed_time ) if $log_is_debug;
     return $obj->value;
 }
 
+sub _record_get_stats {
+    my ( $self, $stat, $elapsed_time ) = @_;
+    $self->{ns_stats}->{$stat}++;
+    $self->{ns_stats}->{'get_time_ms'} += $elapsed_time;
+}
+
 sub unpack_from_data {
     my ( $self, $key, $data ) = @_;
 
-    return CHI::CacheObject->unpack_from_data( $key, $data, $self->serializer );
+    return $self->cache_object_class->unpack_from_data( $key, $data,
+        $self->serializer );
 }
 
 sub get_object {
@@ -299,7 +403,18 @@ sub set {
                 $options = { expires_in => $options };
             }
         }
-        $options = { %{ $self->_default_set_options }, %$options };
+
+        # Disregard default expires_at and expires_in if either are provided
+        #
+        if (   exists( $options->{expires_at} )
+            || exists( $options->{expires_in} ) )
+        {
+            $options =
+              { expires_variance => $self->expires_variance, %$options };
+        }
+        else {
+            $options = { %{ $self->_default_set_options }, %$options };
+        }
     }
 
     $self->set_with_options( $key, $value, $options );
@@ -307,7 +422,10 @@ sub set {
 
 sub set_with_options {
     my ( $self, $key, $value, $options ) = @_;
-    my $ns_stats = $self->{ns_stats};
+    my $ns_stats     = $self->{ns_stats};
+    my $log_is_debug = $log->is_debug;
+    my $measure_time = defined($ns_stats) || $log_is_debug;
+    my ( $start_time, $elapsed_time );
 
     # Determine early and final expiration times
     #
@@ -326,12 +444,16 @@ sub set_with_options {
     # Pack into data, and store
     #
     my $obj =
-      CHI::CacheObject->new( $key, $value, $created_at, $early_expires_at,
-        $expires_at, $self->serializer );
+      $self->cache_object_class->new( $key, $value, $created_at,
+        $early_expires_at, $expires_at, $self->serializer,
+        $self->compress_threshold );
     if ( defined( my $obj_ref = $options->{obj_ref} ) ) {
         $$obj_ref = $obj;
     }
+    $start_time = gettimeofday() if $measure_time;
     if ( $self->set_object( $key, $obj ) ) {
+        $elapsed_time = ( gettimeofday() - $start_time ) * 1000
+          if $measure_time;
 
         # Log the set
         #
@@ -339,9 +461,10 @@ sub set_with_options {
             $ns_stats->{'sets'}++;
             $ns_stats->{'set_key_size'}   += length( $obj->key );
             $ns_stats->{'set_value_size'} += $obj->size;
+            $ns_stats->{'set_time_ms'}    += $elapsed_time;
         }
-        if ( $log->is_debug ) {
-            $self->_log_set_result( $log, $obj );
+        if ($log_is_debug) {
+            $self->_log_set_result( $log, $obj, $elapsed_time );
         }
     }
 
@@ -351,8 +474,14 @@ sub set_with_options {
 sub set_object {
     my ( $self, $key, $obj ) = @_;
 
-    my $data = $obj->pack_to_data();
-    eval { $self->store( $key, $data ) };
+    my $data               = $obj->pack_to_data();
+    my $expires_on_backend = $self->expires_on_backend;
+    my @expires_in         = (
+        $expires_on_backend && $obj->expires_at < CHI_Max_Time
+        ? ( ( $obj->expires_at - time ) * $expires_on_backend )
+        : ()
+    );
+    eval { $self->store( $key, $data, @expires_in ) };
     if ( my $error = $@ ) {
         $self->{ns_stats}->{'set_errors'}++ if defined( $self->{ns_stats} );
         $self->_handle_set_error( $error, $obj );
@@ -369,7 +498,8 @@ sub get_keys_iterator {
 }
 
 sub clear {
-    my ($self) = @_;
+    my $self = shift;
+    die "clear takes no arguments" if @_;
 
     $self->remove_multi( [ $self->get_keys() ] );
 }
@@ -388,23 +518,50 @@ sub expire {
 }
 
 sub compute {
-    my ( $self, $key, $code, $set_options ) = @_;
+    my $self      = shift;
+    my $key       = shift;
+    my $wantarray = wantarray();
+
+    # Allow these in either order for backward compatibility
+    my ( $code, $options ) =
+      ( ref( $_[0] ) eq 'CODE' ) ? ( $_[0], $_[1] ) : ( $_[1], $_[0] );
+
     croak "must specify key and code" unless defined($key) && defined($code);
 
-    my $value = $self->get($key);
+    my %get_options =
+      ( ref($options) eq 'HASH' )
+      ? slice_grep { /(?:expire_if|busy_lock)/ } $options
+      : ();
+    my $set_options =
+        ( ref($options) eq 'HASH' )
+      ? { slice_grep { !/(?:expire_if|busy_lock)/ } $options }
+      : $options;
+
+    my $value = $self->get( $key, %get_options );
     if ( !defined $value ) {
-        $value = $code->();
+        my ( $start_time, $elapsed_time );
+        my $ns_stats = $self->{ns_stats};
+        $start_time = gettimeofday if defined($ns_stats);
+        $value = $wantarray ? [ $code->() ] : $code->();
+        $elapsed_time = ( gettimeofday() - $start_time ) * 1000
+          if defined($ns_stats);
         $self->set( $key, $value, $set_options );
+        if ( defined($ns_stats) ) {
+            $ns_stats->{'computes'}++;
+            $ns_stats->{'compute_time_ms'} += $elapsed_time;
+        }
     }
-    return $value;
+    return $wantarray ? @$value : $value;
 }
 
 sub purge {
     my ($self) = @_;
 
     foreach my $key ( $self->get_keys() ) {
-        if ( $self->get_object($key)->is_expired() ) {
-            $self->remove($key);
+        if ( my $obj = $self->get_object($key) ) {
+            if ( $obj->is_expired() ) {
+                $self->remove($key);
+            }
         }
     }
 }
@@ -427,6 +584,36 @@ sub is_empty {
     return !$self->get_keys();
 }
 
+#
+# (SEMI-) ATOMIC OPERATIONS
+#
+
+sub add {
+    my $self = shift;
+    my $key  = shift;
+
+    if ( !$self->is_valid($key) ) {
+        $self->set( $key, @_ );
+    }
+}
+
+sub append {
+    my ( $self, $key, $new ) = @_;
+
+    my $current = $self->fetch($key) or return undef;
+    $self->store( $key, $current . $new );
+    return 1;
+}
+
+sub replace {
+    my $self = shift;
+    my $key  = shift;
+
+    if ( $self->is_valid($key) ) {
+        $self->set( $key, @_ );
+    }
+}
+
 #
 # MULTI KEY OPERATIONS
 #
@@ -437,18 +624,32 @@ sub fetch_multi_hashref {
     return { map { ( $_, $self->fetch($_) ) } @$keys };
 }
 
+sub get_multi_hashref_objects {
+    my ( $self, $keys ) = @_;
+    my $key_data = $self->fetch_multi_hashref($keys);
+    return {
+        map {
+            my $data = $key_data->{$_};
+            defined($data)
+              ? ( $_, $self->unpack_from_data( $_, $data ) )
+              : ( $_, undef )
+        } keys(%$key_data)
+    };
+}
+
 sub get_multi_arrayref {
     my ( $self, $keys ) = @_;
     croak "must specify keys" unless defined($keys);
     my $transformed_keys = [ map { $self->transform_key($_) } @$keys ];
 
     my $key_count = scalar(@$keys);
-    my $keyvals   = $self->fetch_multi_hashref($transformed_keys);
+    my $keyobjs   = $self->get_multi_hashref_objects($transformed_keys);
     return [
         map {
-            $self->get( $keys->[$_],
-                data => $keyvals->{ $transformed_keys->[$_] } )
-          } ( 0 .. $key_count - 1 )
+            my $key = $transformed_keys->[$_];
+            my $obj = $keyobjs->{$key};
+            $obj ? $self->get( $key, obj => $obj ) : undef
+        } ( 0 .. $key_count - 1 )
     ];
 }
 
@@ -462,6 +663,11 @@ sub get_multi_hashref {
 }
 
 sub set_multi {
+    my $self = shift;
+    $self->store_multi(@_);
+}
+
+sub store_multi {
     my ( $self, $key_values, $set_options ) = @_;
     croak "must specify key_values" unless defined($key_values);
 
@@ -599,14 +805,17 @@ sub _dispatch_error_msg {
 }
 
 sub _describe_cache_get {
-    my ( $self, $key ) = @_;
+    my ( $self, $key, $elapsed_time ) = @_;
 
-    return sprintf( "cache get for namespace='%s', key='%s', cache='%s'",
-        $self->namespace, $key, $self->label );
+    return
+      sprintf( "cache get for namespace='%s', key='%s', cache='%s'"
+          . ( defined($elapsed_time) ? ", time='%dms'" : "" ),
+        $self->namespace, $key, $self->label,
+        defined($elapsed_time) ? int($elapsed_time) : () );
 }
 
 sub _describe_cache_set {
-    my ( $self, $obj ) = @_;
+    my ( $self, $obj, $elapsed_time ) = @_;
 
     my $expires_str = (
         ( $obj->expires_at == CHI_Max_Time )
@@ -620,8 +829,11 @@ sub _describe_cache_set {
 
     return
       sprintf(
-        "cache set for namespace='%s', key='%s', size=%d, expires='%s', cache='%s'",
-        $self->namespace, $obj->key, $obj->size, $expires_str, $self->label );
+        "cache set for namespace='%s', key='%s', size=%d, expires='%s', cache='%s'"
+          . ( defined($elapsed_time) ? ", time='%dms'" : "" ),
+        $self->namespace, $obj->key, $obj->size, $expires_str, $self->label,
+        defined($elapsed_time) ? int($elapsed_time) : () );
+
 }
 
 1;
@@ -632,7 +844,11 @@ __END__
 
 =head1 NAME
 
-CHI::Driver -- Base class for all CHI drivers.
+CHI::Driver - Base class for all CHI drivers
+
+=head1 VERSION
+
+version 0.58
 
 =head1 DESCRIPTION
 
@@ -643,15 +859,19 @@ See L<CHI/METHODS> for documentation on $cache methods, and
 L<CHI::Driver::Development|CHI::Driver::Development> for documentation on
 creating new subclasses of CHI::Driver.
 
+=head1 SEE ALSO
+
+L<CHI|CHI>
+
 =head1 AUTHOR
 
-Jonathan Swartz
+Jonathan Swartz <swartz@pobox.com>
 
-=head1 COPYRIGHT & LICENSE
+=head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2007 Jonathan Swartz.
+This software is copyright (c) 2012 by Jonathan Swartz.
 
-This program is free software; you can redistribute it and/or modify it under
-the same terms as Perl itself.
+This is free software; you can redistribute it and/or modify it under
+the same terms as the Perl 5 programming language system itself.
 
 =cut
